from django.db import models
from django.db.models import Sum, Max, Min, Count, Avg
from django.db.models.functions import TruncMonth, TruncQuarter, TruncYear, TruncWeek, TruncDay

NUMBER_FIELDS = (models.IntegerField, models.PositiveSmallIntegerField, models.PositiveIntegerField)
DATE_FIELDS = (models.DateTimeField, models.DateField)

ANNOTATION_VALUE_CHOICES = (('', ''),
                            ('y', 'Year'),
                            ('q', 'Quarter'),
                            ('m', 'Month'),
                            ('w', 'Week'),
                            ('d', 'Day'))

ANNOTATION_VALUE_FUNCTIONS = {'y': TruncYear,
                              'q': TruncQuarter,
                              'm': TruncMonth,
                              'w': TruncWeek,
                              'd': TruncDay}

ANNOTATIONS_CHOICES = (('', ''),
                       ('sum', 'Sum'),
                       ('max', 'Max'),
                       ('min', 'Min'),
                       ('count', 'Count'),
                       ('avg', 'Avg'))

ANNOTATION_FUNCTIONS = {'sum': Sum,
                        'max': Max,
                        'min': Min,
                        'count': Count,
                        'avg': Avg}

DATE_FORMAT_TYPE_DD_MM_YY_SLASH = 1
DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH = 2
DATE_FORMAT_TYPE_MM_DD_YY_SLASH = 3
DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH = 4
DATE_FORMAT_TYPE_DD_MM_YY_DASH = 5
DATE_FORMAT_TYPE_DD_MM_YYYY_DASH = 6
DATE_FORMAT_TYPE_MM_DD_YY_DASH = 7
DATE_FORMAT_TYPE_MM_DD_YYYY_DASH = 8
DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY = 9
DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY = 10
DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY = 11
DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY = 12
DATE_FORMAT_TYPE_YYYY = 13
DATE_FORMAT_TYPE_YY = 14
DATE_FORMAT_TYPE_MM_YY = 15
DATE_FORMAT_TYPE_MM_YYYY = 16
DATE_FORMAT_TYPE_WORDS_MM_YY = 17
DATE_FORMAT_TYPE_WORDS_MM_YYYY = 18
DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY = 19
DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY = 20
DATE_FORMAT_TYPE_MM = 21
DATE_FORMAT_TYPE_WORDS_MM = 22
DATE_FORMAT_TYPE_SHORT_WORDS_MM = 23
DATE_FORMAT_TYPE_WW = 24
DATE_FORMAT_TYPE_WW_YYYY_DASH = 25
DATE_FORMAT_TYPE_WW_YY_DASH = 26


DATE_FORMAT_TYPES_DJANGO_FORMAT = {DATE_FORMAT_TYPE_DD_MM_YY_SLASH: '%d/%m/%y',
                                   DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH: '%d/%m/%Y',
                                   DATE_FORMAT_TYPE_MM_DD_YY_SLASH: '%m/%d/%y',
                                   DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH: '%m/%d/%Y',
                                   DATE_FORMAT_TYPE_DD_MM_YY_DASH: '%d-%m-%y',
                                   DATE_FORMAT_TYPE_DD_MM_YYYY_DASH: '%d-%m-%Y',
                                   DATE_FORMAT_TYPE_MM_DD_YY_DASH: '%m-%d-%y',
                                   DATE_FORMAT_TYPE_MM_DD_YYYY_DASH: '%m-%d-%Y',
                                   DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY: '%B %d %Y',
                                   DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY: '%d %B %Y',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY: '%b %d %Y',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY: '%d %b %Y',
                                   DATE_FORMAT_TYPE_YYYY: '%Y',
                                   DATE_FORMAT_TYPE_YY: '%y',
                                   DATE_FORMAT_TYPE_MM_YY: '%m %y',
                                   DATE_FORMAT_TYPE_MM_YYYY: '%m %Y',
                                   DATE_FORMAT_TYPE_WORDS_MM_YY: '%B %y',
                                   DATE_FORMAT_TYPE_WORDS_MM_YYYY: '%B %Y',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY: '%b %y',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY: '%b %Y',
                                   DATE_FORMAT_TYPE_MM: '%m',
                                   DATE_FORMAT_TYPE_WORDS_MM: '%B',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_MM: '%b',
                                   DATE_FORMAT_TYPE_WW: '%W',
                                   DATE_FORMAT_TYPE_WW_YYYY_DASH: '%W-%Y',
                                   DATE_FORMAT_TYPE_WW_YY_DASH: '%W-%y',

                                   }

DATE_FORMAT_TYPES = (
    ('', ''),
    (DATE_FORMAT_TYPE_DD_MM_YY_SLASH, 'dd/mm/yy'),
    (DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH, 'dd/mm/yyyy'),
    (DATE_FORMAT_TYPE_MM_DD_YY_SLASH, 'mm/dd/yy'),
    (DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH, 'mm/dd/yyyy'),
    (DATE_FORMAT_TYPE_DD_MM_YY_DASH, 'dd-mm-yy'),
    (DATE_FORMAT_TYPE_DD_MM_YYYY_DASH, 'dd-mm-yyyy'),
    (DATE_FORMAT_TYPE_MM_DD_YY_DASH, 'mm-dd-yy'),
    (DATE_FORMAT_TYPE_MM_DD_YYYY_DASH, 'mm-dd-yyyy'),
    (DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY, 'MM d yyyy'),
    (DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY, 'd MM yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY, 'M d yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY, 'd M yyyy'),
    (DATE_FORMAT_TYPE_YYYY, 'yyyy'),
    (DATE_FORMAT_TYPE_YY, 'yy'),
    (DATE_FORMAT_TYPE_MM_YY, 'mm yy'),
    (DATE_FORMAT_TYPE_MM_YYYY, 'mm yyyy'),
    (DATE_FORMAT_TYPE_WORDS_MM_YY, 'MM yy'),
    (DATE_FORMAT_TYPE_WORDS_MM_YYYY, 'MM yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY, 'b yy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY, 'b yyyy'),
    (DATE_FORMAT_TYPE_MM, 'mm'),
    (DATE_FORMAT_TYPE_WORDS_MM, 'MM'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM, 'd'),
    (DATE_FORMAT_TYPE_WW, 'WW'),
    (DATE_FORMAT_TYPE_WW_YYYY_DASH, 'WW-YYYY'),
    (DATE_FORMAT_TYPE_WW_YY_DASH, 'WW-YY'),
)



from django.contrib import admin
from advanced_report_builder.models import Report, ReportType, TableReport, ReportQuery


@admin.register(ReportQuery)
class ReportQueryAdmin(admin.ModelAdmin):
    model = ReportQuery
    list_display = ('report',
                    'name',
                    'query')


class ReportQueryInline(admin.TabularInline):
    model = ReportQuery


@admin.register(Report)
class ReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'instance_type',
                    )
    inlines = [ReportQueryInline]


@admin.register(ReportType)
class ReportTypeAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'content_type',
                    'report_builder_class_name')


@admin.register(TableReport)
class ReportTableAdmin(admin.ModelAdmin):
    list_display = ('name',
                    )
    exclude = ['instance_type',
               ]
    inlines = [ReportQueryInline]







from ajax_helpers.mixins import AjaxHelpers
from django.shortcuts import get_object_or_404
from django.views.generic import TemplateView
from django_menus.menu import MenuMixin

from advanced_report_builder.models import Report
from advanced_report_builder.utils import split_slug
from advanced_report_builder.views.datatables import TableView


class ViewReportBase(AjaxHelpers, MenuMixin, TemplateView):
    model = Report
    views = {'tablereport': TableView}

    def get_object(self):
        slug = split_slug(self.kwargs['slug'])
        return get_object_or_404(self.model, pk=slug['pk'])

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        view = self.get_view(report=self.get_object())
        report_data = view.as_view()(self.request, *self.args, **self.kwargs).rendered_content
        context['report'] = report_data
        return context

    def get_view(self, report):
        return self.views.get(report.instance_type)

    def post(self, request, *args, **kwargs):
        view = self.get_view(report=self.get_object())
        return view.as_view()(self.request, *self.args, **self.kwargs)




# version = pip_version('django-advance-query-builder')
from ajax_helpers.html_include import SourceBase


class ChartJS(SourceBase):
    cdn_path = 'https://cdnjs.cloudflare.com/ajax/libs/Chart.js/3.5.1/chart.min.js'


class Dot(SourceBase):
    static_path = '/report_builder/dot/'
    js_filename = 'doT.js'


class JQueryExtendext(SourceBase):
    static_path = '/report_builder/jquery_extendext/'
    js_filename = 'jQuery.extendext.js'


class QueryBuilder(SourceBase):
    static_path = '/report_builder/query_builder/'
    css_filename = 'query-builder.default.css'
    js_filename = 'query-builder.min.js'


packages = {
    'query_builder': [JQueryExtendext, Dot, QueryBuilder],
}



import base64
import json

from django.forms import CharField, ChoiceField, BooleanField, IntegerField
from django.shortcuts import get_object_or_404
from django_datatables.datatables import ColumnInitialisor
from django_modals.forms import CrispyForm
from django_modals.widgets.widgets import Toggle

from advanced_report_builder.globals import DATE_FIELDS, NUMBER_FIELDS, ANNOTATION_VALUE_CHOICES, ANNOTATIONS_CHOICES, \
    DATE_FORMAT_TYPES
from advanced_report_builder.models import ReportType


from advanced_report_builder.utils import split_attr


class FieldForm(CrispyForm):

    def get_django_field(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        report_type = get_object_or_404(ReportType, pk=self.slug['report_type_id'])
        base_model = report_type.content_type.model_class()

        column_initialisor = ColumnInitialisor(start_model=base_model, path=data['field'])
        cols = column_initialisor.get_columns()

        if column_initialisor.django_field is None and cols:
            column_initialisor = ColumnInitialisor(start_model=base_model, path=cols[0].field)
            column_initialisor.get_columns()

        return column_initialisor.django_field

    def setup_modal(self, *args, **kwargs):
        data = json.loads(base64.b64decode(self.slug['data']))
        django_field = self.get_django_field()

        self.fields['title'] = CharField(initial=data['title'])

        if django_field is not None:
            data_attr = split_attr(data)

            if isinstance(django_field, DATE_FIELDS):

                self.fields['annotations_value'] = ChoiceField(choices=ANNOTATION_VALUE_CHOICES, required=False)
                if 'annotations_value' in data_attr:
                    self.fields['annotations_value'].initial = data_attr['annotations_value']

                self.fields['date_format'] = ChoiceField(choices=DATE_FORMAT_TYPES, required=False)
                if 'date_format' in data_attr:
                    self.fields['date_format'].initial = data_attr['date_format']

            elif isinstance(django_field, NUMBER_FIELDS):
                self.fields['annotations_type'] = ChoiceField(choices=ANNOTATIONS_CHOICES, required=False)
                if 'annotations_type' in data_attr:
                    self.fields['annotations_type'].initial = data_attr['annotations_type']

                self.fields['show_totals'] = BooleanField(required=False,
                                                          widget=Toggle(attrs={'data-onstyle': 'success',
                                                                               'data-on': 'YES',
                                                                               'data-off': 'NO'}))
                if 'show_totals' in data_attr and data_attr['show_totals'] == '1':
                    self.fields['show_totals'].initial = True

                self.fields['decimal_places'] = IntegerField()
                self.fields['decimal_places'].initial = int(data_attr.get('decimal_places', 0))

        super().setup_modal(*args, **kwargs)

    def get_additional_attributes(self):
        attributes = []
        django_field = self.get_django_field()
        if django_field is not None:
            if isinstance(django_field, DATE_FIELDS):
                if self.cleaned_data['annotations_value']:
                    attributes.append(f'annotations_value-{self.cleaned_data["annotations_value"]}')
                if self.cleaned_data['date_format']:
                    attributes.append(f'date_format-{self.cleaned_data["date_format"]}')

            elif isinstance(django_field, NUMBER_FIELDS):
                if self.cleaned_data['annotations_type']:
                    attributes.append(f'annotations_type-{self.cleaned_data["annotations_type"]}')
                if self.cleaned_data['show_totals'] and self.cleaned_data["show_totals"]:
                    attributes.append('show_totals-1')
                if self.cleaned_data['decimal_places'] > 0:
                    attributes.append(f'decimal_places-{self.cleaned_data["decimal_places"]}')

        if attributes:
            return '-'.join(attributes)

        return None



import json
import operator
from functools import reduce

from django.db.models import Q

from advanced_report_builder.variable_date import VariableDate


class FilterQueryMixin:

    def process_filters(self, query, search_filter_data):
        if not search_filter_data:
            return query

        query_data = json.loads(search_filter_data)
        query_list = self._process_group(query_data=query_data)
        reduce_by = self._format_group_conditions(query_data['condition'])

        if query_list:
            return query.filter(reduce(reduce_by, query_list))

        return query

    @staticmethod
    def _get_operator(display_operator):
        """
        Converts the display operators into operators for the Q object.
        :param display_operator:
        :return:
        """

        operators_dict = {
            'equal': '',
            'not_equal': '',  # Not equals is now done by inverting the individual Q object. (for dates).
            'less': '__lt',
            'less_or_equal': '__lte',
            'greater': '__gt',
            'greater_or_equal': '__gte',
            'contains': '__icontains',
            'not_contains': '__icontains',
            'begins_with': '__istartswith',
            'not_begins_with': '__istartswith',
            'ends_with': '__iendswith',
            'not_ends_with': '__iendswith',
            'in': '__in',
            'not_in': '__in',
            'is_null': '__isnull',
            'is_not_null': '__isnull',
        }

        query_operator = operators_dict.get(display_operator)
        return query_operator

    @staticmethod
    def _format_group_conditions(display_condition):
        """
        Converts the display condition from the javascript into a condition for the Q object.
        :param display_condition:
        :return:
        """
        if display_condition == 'AND':
            reduce_by = operator.and_
        else:
            reduce_by = operator.or_

        return reduce_by

    def _process_group(self, query_data):
        query_list = []

        for rule in query_data['rules']:
            if condition := rule.get('condition'):
                reduce_by = self._format_group_conditions(condition)
                sub_query_list = self._process_group(query_data=rule)
                if sub_query_list:
                    query_list.append(reduce(reduce_by, sub_query_list))
                continue

            field = rule['field']
            _id = rule['id']
            display_operator = rule['operator']
            query_string = field + self._get_operator(display_operator)
            data_type = rule['type']
            value = rule.get('value', '')
            if display_operator == 'is_null':
                value = True
            elif display_operator == 'is_not_null':
                value = False

            if data_type == "string" and _id.endswith('__variable_date'):
                self.get_variable_date(value=value,
                                       query_list=query_list,
                                       display_operator=display_operator,
                                       field=field,
                                       query_string=query_string)
            else:
                # 'Normal' Query.
                if display_operator in ["not_equal", "not_in", "not_contains", 'not_begins_with', "not_ends_with"]:
                    query_list.append(~Q((query_string, value)))
                else:
                    query_list.append(Q((query_string, value)))

            # elif data_type == 'date':
            #     value = datetime.strptime(value, self.date_format).date()
            # elif data_type == 'datetime':
            #     value = datetime.strptime(value, self.date_time_format)
            # elif data_type == 'integer' and value == '#current_user#':
            #     value = self.user_profile_id

            variable_date = False
        return query_list

    @staticmethod
    def get_variable_date(value, query_list, display_operator, field, query_string):
        _, range_type = value.split(":")
        variable_date = VariableDate()
        value = variable_date.get_variable_dates(range_type=int(range_type))
        if display_operator in ['less', 'greater_or_equal']:
            query_list.append(Q((query_string, value[0])))
        elif display_operator in ['greater', 'less_or_equal']:
            query_list.append(Q((query_string, value[1])))
        elif display_operator in ["not_equal", "not_in"]:
            query_list.append(~((Q((field + "__gte", value[0]))) & (Q((field + "__lte", value[1])))))
        else:
            query_list.append(((Q((field + "__gte", value[0]))) & (Q((field + "__lte", value[1])))))



# Generated by Django 3.2.7 on 2021-11-29 11:24

from django.db import migrations, models
import django.db.models.deletion
import time_stamped_model.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('instance_type', models.CharField(max_length=255, null=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='TableReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('table_fields', models.TextField(blank=True, null=True)),
                ('has_clickable_rows', models.BooleanField(default=False)),
                ('pivot_fields', models.TextField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.CreateModel(
            name='ReportType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('report_builder_class_name', models.CharField(max_length=200)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='contenttypes.contenttype')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.AddField(
            model_name='report',
            name='report_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype'),
        ),
        migrations.CreateModel(
            name='ReportQuery',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.TextField(default='Standard')),
                ('query', models.JSONField(blank=True, null=True)),
                ('report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.report')),
            ],
            options={
                'verbose_name_plural': 'Report queries',
                'ordering': ['name'],
                'unique_together': {('name', 'report')},
            },
        ),
    ]



import base64
import copy
import json

from ajax_helpers.mixins import AjaxHelpers
from crispy_forms.bootstrap import StrictButton
from django.apps import apps
from django.forms import CharField
from django.shortcuts import get_object_or_404
from django_datatables.datatables import DatatableView, ColumnInitialisor
from django_datatables.plugins.column_totals import ColumnTotals
from django_menus.menu import MenuMixin, MenuItem
from django_modals.fields import FieldEx
from django_modals.forms import ModelCrispyForm
from django_modals.modals import ModelFormModal, FormModal
from django_modals.widgets.widgets import Toggle

from advanced_report_builder.columns import ReportBuilderDateColumn, ReportBuilderNumberColumn
from advanced_report_builder.field_types import FieldTypes
from advanced_report_builder.filter_query import FilterQueryMixin
from advanced_report_builder.forms import FieldForm
from advanced_report_builder.globals import DATE_FORMAT_TYPES_DJANGO_FORMAT, ANNOTATION_VALUE_FUNCTIONS, DATE_FIELDS, \
    ANNOTATION_FUNCTIONS, NUMBER_FIELDS
from advanced_report_builder.models import TableReport, ReportType, ReportQuery
from advanced_report_builder.utils import split_attr, split_slug


class TableView(AjaxHelpers, FilterQueryMixin, MenuMixin, DatatableView):
    template_name = 'advanced_report_builder/datatable.html'

    date_field = ReportBuilderDateColumn
    number_field = ReportBuilderNumberColumn

    def __init__(self, *args, **kwargs):
        self.table_report = None
        self.slug = None
        super().__init__(*args, **kwargs)

    def add_tables(self):
        return None

    def dispatch(self, request, *args, **kwargs):
        self.slug = split_slug(self.kwargs['slug'])
        self.table_report = get_object_or_404(TableReport, pk=self.slug['pk'])
        base_model = self.table_report.get_base_modal()
        self.add_table(type(self).__name__.lower(), model=base_model)
        return super().dispatch(request, *args, **kwargs)

    def get_date_field(self, table_field, fields):
        data_attr = split_attr(table_field)
        field_name = table_field['field']
        date_format = data_attr.get('date_format')
        if date_format:
            date_format = DATE_FORMAT_TYPES_DJANGO_FORMAT[int(date_format)]

        date_function_kwargs = {'title': table_field.get('title'),
                                'date_format': date_format}

        annotations_value = data_attr.get('annotations_value')
        if annotations_value:
            new_field_name = f'{annotations_value}_{field_name}'
            function = ANNOTATION_VALUE_FUNCTIONS[annotations_value]
            date_function_kwargs['annotations_value'] = {new_field_name: function(field_name)}
            field_name = new_field_name

        date_function_kwargs.update({'field': field_name,
                                     'column_name': field_name})

        field = self.date_field(**date_function_kwargs)
        fields.append(field)
        return field_name

    def get_number_field(self, table_field, fields, totals, col_type_override):
        data_attr = split_attr(table_field)
        field_name = table_field['field']
        css_class = None

        annotations_type = data_attr.get('annotations_type')

        if col_type_override:
            field = copy.deepcopy(col_type_override)
            title = table_field.get('title')
            if annotations_type == 'count':
                new_field_name = f'{annotations_type}_{field_name}'
                number_function_kwargs = {}
                if title:
                    number_function_kwargs['title'] = title
                function = ANNOTATION_FUNCTIONS[annotations_type]
                number_function_kwargs['annotations'] = {new_field_name: function(field.field)}

                number_function_kwargs.update({'field': new_field_name,
                                               'column_name': field_name})
                field = self.number_field(**number_function_kwargs)
            else:
                css_class = field.column_defs.get('className')
                if title:
                    field.title = title
                if annotations_type:
                    new_field_name = f'{annotations_type}_{field_name}'

                    function = ANNOTATION_FUNCTIONS[annotations_type]
                    field.annotations = {new_field_name: function(field.field)}
                    field.field = new_field_name

            fields.append(field)
        else:
            number_function_kwargs = {'title': table_field.get('title')}
            decimal_places = data_attr.get('decimal_places')

            if decimal_places:
                number_function_kwargs = {'decimal_places': int(decimal_places)}

            if annotations_type:
                new_field_name = f'{annotations_type}_{field_name}'
                function = ANNOTATION_FUNCTIONS[annotations_type]
                number_function_kwargs['annotations'] = {new_field_name: function(field_name)}
                field_name = new_field_name

            number_function_kwargs.update({'field': field_name,
                                           'column_name': field_name})

            field = self.number_field(**number_function_kwargs)
            fields.append(field)

        decimal_places = data_attr.get('decimal_places', 0)
        show_total = data_attr.get('show_totals')
        if show_total == '1':
            totals[field_name] = {'sum': 'to_fixed', 'decimal_places': decimal_places, 'css_class': css_class}

        return field_name

    def setup_table(self, table):
        table.extra_filters = self.extra_filters
        table_fields = json.loads(self.table_report.table_fields)
        fields = []
        totals = {}
        base_modal = self.table_report.get_base_modal()
        first_field_name = None
        for table_field in table_fields:
            field = table_field['field']

            original_column_initialisor = ColumnInitialisor(start_model=base_modal, path=field)
            cols = original_column_initialisor.get_columns()
            django_field = original_column_initialisor.django_field
            col_type_override = None

            if django_field is None and cols:
                col_type_override = cols[0]
                column_initialisor = ColumnInitialisor(start_model=base_modal, path=col_type_override.field)
                column_initialisor.get_columns()
                django_field = column_initialisor.django_field

            if isinstance(django_field, DATE_FIELDS):
                field_name = self.get_date_field(table_field=table_field, fields=fields)

            elif isinstance(django_field, NUMBER_FIELDS):
                field_name = self.get_number_field(table_field=table_field,
                                                   fields=fields,
                                                   totals=totals,
                                                   col_type_override=col_type_override)
            else:
                field_attr = {}
                if 'title' in table_field:
                    field_attr['title'] = table_field['title']
                field_name = field
                if field_attr:
                    field = (field, field_attr)
                fields.append(field)
            if not first_field_name:
                first_field_name = field_name
        table.add_columns(*fields)
        if totals:
            totals[first_field_name] = {'text': 'Totals'}
            table.add_plugin(ColumnTotals, totals)

    def get_report_query(self):
        query_id = self.slug.get(f'query{self.table_report.id}')
        if query_id:
            report_query = get_object_or_404(ReportQuery, id=query_id)
            if report_query.report_id != self.table_report.id:
                return None
        else:
            report_query = self.table_report.reportquery_set.first()
        return report_query

    def extra_filters(self, query):
        report_query = self.get_report_query()
        if not report_query:
            return query

        return self.process_filters(query=query,
                                    search_filter_data=report_query.query)

    def add_to_context(self, **kwargs):
        return {'table_report': self.table_report}

    def setup_menu(self):
        super().setup_menu()
        self.add_menu('button_menu', 'button_group').add_items(
            *self.pod_menu(),
            *self.queries_menu(),
        )

    def pod_menu(self):

        query_id = self.slug.get(f'query{self.table_report.id}')
        slug_str = ''
        if query_id:
            slug_str = f'-query_id-{query_id}'

        return [MenuItem(f'advanced_report_builder:table_modal,pk-{self.table_report.id}{slug_str}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary'])]

    def queries_menu(self):
        return []


class ReportBaseForm(ModelCrispyForm):
    submit_class = 'btn-success modal-submit'

    def submit_button(self, css_class=submit_class, button_text='Submit', **kwargs):
        return StrictButton(button_text, css_class=css_class)


class TableModal(ModelFormModal):
    size = 'xl'
    model = TableReport
    base_form = ReportBaseForm
    ajax_commands = ['button', 'select2', 'ajax']

    form_fields = ['name',
                   ('has_clickable_rows', {'widget': Toggle(attrs={'data-onstyle': 'success',
                                                                   'data-on': 'YES',
                                                                   'data-off': 'NO'})}),
                   'report_type',
                   'table_fields',
                   ]

    def __init__(self, *args, **kwargs):
        self.report_query = None
        self.show_query_name = False
        super().__init__(*args, **kwargs)

    def form_setup(self, form, *_args, **_kwargs):
        self.report_query = None
        form.fields['query_data'] = CharField(required=False)

        if self.object.id:
            query_id = self.slug.get('query_id')
            if query_id:
                self.report_query = get_object_or_404(ReportQuery, id=query_id)
            else:
                self.report_query = self.object.reportquery_set.first()

            if self.report_query:
                self.show_query_name = self.object.reportquery_set.count() > 1
                if self.show_query_name:
                    form.fields['query_name'] = CharField(required=True, initial=self.report_query.name)

                form.fields['query_data'].initial = self.report_query.query

        fields = [FieldEx('name'),
                  FieldEx('report_type'),
                  FieldEx('has_clickable_rows', template='django_modals/fields/label_checkbox.html'),
                  FieldEx('table_fields', template='advanced_report_builder/fields/select_column.html')]

        if self.show_query_name:
            fields.append(FieldEx('query_name'))

        fields.append(FieldEx('query_data', template='advanced_report_builder/fields/query_builder.html'))
        return fields

    def form_valid(self, form):
        # Create a item using some of the extra fields and save it, then attach it to the SiteVisit.
        table_report = form.save()

        if not self.report_query and form.cleaned_data['query_data']:
            ReportQuery(query=form.cleaned_data['query_data'],
                        report=table_report).save()
        elif form.cleaned_data['query_data']:
            self.report_query.query = form.cleaned_data['query_data']
            if self.show_query_name:
                self.report_query.name = form.cleaned_data['query_name']
            self.report_query.save()
        elif self.report_query:
            self.report_query.delete()

        return self.command_response('reload')

    def _get_fields(self, base_model, fields, tables, report_builder_fields,
                    prefix='', title_prefix='', title=None, colour=None):
        if title is None:
            title = report_builder_fields.title
        if colour is None:
            colour = report_builder_fields.colour

        tables.append({'name': title,
                       'colour': colour})

        for report_builder_field in report_builder_fields.fields:

            column_initialisor = ColumnInitialisor(start_model=base_model, path=report_builder_field)
            columns = column_initialisor.get_columns()
            for column in columns:
                fields.append({'field': prefix + column.column_name,
                               'label': title_prefix + column.title,
                               'colour': report_builder_fields.colour})

        for include in report_builder_fields.includes:
            app_label, model, report_builder_fields_str = include['model'].split('.')
            new_model = apps.get_model(app_label, model)
            new_report_builder_fields = getattr(new_model, report_builder_fields_str, None)
            self._get_fields(base_model=new_model,
                             fields=fields,
                             tables=tables,
                             report_builder_fields=new_report_builder_fields,
                             prefix=f"{include['field']}__",
                             title_prefix=include['title'] + ' -> ',
                             title=include.get('title'),
                             colour=include.get('colour'))

    def _get_query_builder_fields(self, base_model, query_builder_filters, report_builder_fields, prefix='',
                                  title_prefix=''):

        field_types = FieldTypes()

        for report_builder_field in report_builder_fields.fields:

            column_initialisor = ColumnInitialisor(start_model=base_model, path=report_builder_field)
            columns = column_initialisor.get_columns()
            for column in columns:
                if column_initialisor.django_field is not None:
                    field_types.get_filter(query_builder_filters=query_builder_filters,
                                           django_field=column_initialisor.django_field,
                                           field=prefix + column.column_name,
                                           title=title_prefix + column.title)
        for include in report_builder_fields.includes:
            app_label, model, report_builder_fields_str = include['model'].split('.')
            new_model = apps.get_model(app_label, model)
            new_report_builder_fields = getattr(new_model, report_builder_fields_str, None)
            foreign_key_field = getattr(base_model, include['field'], None).field
            if foreign_key_field.null:
                field_types.get_foreign_key_null_field(query_builder_filters=query_builder_filters,
                                                       field=prefix + include['field'],
                                                       title=title_prefix + include['title'])

            self._get_query_builder_fields(base_model=new_model,
                                           query_builder_filters=query_builder_filters,
                                           report_builder_fields=new_report_builder_fields,
                                           prefix=f"{include['field']}__",
                                           title_prefix=f"{include['title']} --> ")

    def ajax_get_fields(self, **kwargs):
        report_type_id = kwargs['report_type'][0]

        report_type = get_object_or_404(ReportType, pk=report_type_id)
        base_model = report_type.content_type.model_class()
        report_builder_fields = getattr(base_model, report_type.report_builder_class_name, None)
        fields = []
        tables = []
        self._get_fields(base_model=base_model,
                         fields=fields,
                         tables=tables,
                         report_builder_fields=report_builder_fields)
        return self.command_response('report_fields', data=json.dumps({'fields': fields,
                                                                       'tables': tables}))

    def ajax_get_query_builder_fields(self, **kwargs):
        report_type_id = kwargs['report_type'][0]
        report_type = get_object_or_404(ReportType, pk=report_type_id)
        base_model = report_type.content_type.model_class()
        report_builder_fields = getattr(base_model, report_type.report_builder_class_name, None)
        query_builder_filters = []
        self._get_query_builder_fields(base_model=base_model,
                                       query_builder_filters=query_builder_filters,
                                       report_builder_fields=report_builder_fields)

        return self.command_response('query_builder', data=json.dumps(query_builder_filters))


class FieldModal(FormModal):
    form_class = FieldForm

    @property
    def modal_title(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        return f'Edit {data["title"]}'

    def form_valid(self, form):
        selector = self.slug['selector']

        _attr = form.get_additional_attributes()
        self.add_command({'function': 'set_attr',
                          'selector': f'#{selector}',
                          'attr': 'data-attr',
                          'val': _attr})

        self.add_command({'function': 'html', 'selector': f'#{selector} span', 'html': form.cleaned_data['title']})
        self.add_command({'function': 'save_query_builder'})
        self.add_command({'function': 'update_selection'})
        return self.command_response('close')



import json

from django.contrib.contenttypes.models import ContentType
from django.db import models
from time_stamped_model.models import TimeStampedModel


class ReportType(TimeStampedModel):
    name = models.CharField(max_length=200)
    content_type = models.ForeignKey(ContentType, null=False, blank=False, on_delete=models.PROTECT)
    report_builder_class_name = models.CharField(max_length=200)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']


class Report(TimeStampedModel):
    name = models.CharField(max_length=200)
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    instance_type = models.CharField(null=True, max_length=255)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']

    def get_report(self):
        return getattr(self, self.instance_type)

    def get_title(self):
        return self.name

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.instance_type is None:
            self.instance_type = self._meta.label_lower.split('.')[1]
        return super().save(force_insert=force_insert, force_update=force_update,
                            using=using, update_fields=update_fields)

    def get_base_modal(self):
        return self.report_type.content_type.model_class()


class ReportQuery(TimeStampedModel):
    report = models.ForeignKey(Report, on_delete=models.CASCADE)
    name = models.TextField(default='Standard')
    query = models.JSONField(null=True, blank=True)

    class Meta:
        ordering = ['name']
        unique_together = ('name', 'report')
        verbose_name_plural = 'Report queries'


class TableReport(Report):
    table_fields = models.TextField(null=True, blank=True)
    has_clickable_rows = models.BooleanField(default=False)  # This is for standard tables.
    pivot_fields = models.TextField(null=True, blank=True)

    def has_pivot_data(self):
        return self.pivot_fields is not None and self.pivot_fields != ''

    def get_pivot_fields(self):
        pivot_data = json.loads(self.pivot_fields)
        return pivot_data



from django_datatables.columns import ColumnBase


class ReportBuilderDateColumn(ColumnBase):

    def __init__(self, *,  date_format=None, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        if date_format:
            self.date_format = date_format
        else:
            self.date_format = '%d/%m/%Y'

    def row_result(self, data, _page_data):
        try:
            date = data[self.field].strftime(self.date_format)
            return date
        except AttributeError:
            return ""


class ReportBuilderNumberColumn(ColumnBase):

    def __init__(self, *,  decimal_places=0, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        self.decimal_places = f'{{:.{decimal_places}f}}'

    def row_result(self, data, _page_data):
        try:
            number = self.decimal_places.format(data[self.field])
            return number
        except AttributeError:
            return ""



from django.urls import path

from advanced_report_builder.views.datatables import TableModal, FieldModal

app_name = 'report_builder'


urlpatterns = [
    path('table/modal/<str:slug>/', TableModal.as_view(), name='table_modal'),
    path('table/modal/field/<str:slug>/', FieldModal.as_view(), name='field_modal'),
]



from django.conf import settings
from django.utils.module_loading import import_string


def split_attr(data):
    if 'data_attr' not in data:
        return {}
    _attr = {}
    s = data['data_attr'].split('-')
    _attr.update({s[k]: s[k + 1] for k in range(0, int(len(s) - 1), 2)})
    return _attr


def split_slug(slug_str):
    s = slug_str.split('-')
    slug = {}
    if len(s) == 1:
        slug['pk'] = s[0]
    else:
        slug.update({s[k]: s[k + 1] for k in range(0, int(len(s) - 1), 2)})
    return slug


def make_slug_str(slug, overrides=None):
    if len(slug) == 1 and 'pk' in slug and not overrides:
        return slug['pk']

    if not overrides:
        slug_parts = [f'{k}-{v}' for k, v in slug.items()]
    else:
        slug_parts = []
        for key, value in slug.items():
            if key in overrides:
                continue
            slug_parts.append(f'{key}-{value}')
        for key, value in overrides.items():
            slug_parts.append(f'{key}-{value}')

    return '-'.join(slug_parts)


def get_custom_report_builder():
    return import_string(getattr(settings,
                                 'REPORT_BUILDER_CUSTOMISATION',
                                 'report_builder.customise.CustomiseReportBuilder'))
